@@ -1,6 +1,6 @@
 #!/usr/bin/env python3
 
-# SPDX-FileCopyrightText: © 2022 Decompollaborate
+# SPDX-FileCopyrightText: © 2022-2024 Decompollaborate
 # SPDX-License-Identifier: MIT
 
 from __future__ import annotations
@@ -18,7 +18,7 @@ class ElementBase:
     """Represents the base class used for most file sections and symbols.
     """
 
-    def __init__(self, context: Context, vromStart: int, vromEnd: int, inFileOffset: int, vram: int, name: str, words: list[int], sectionType: FileSectionType, segmentVromStart: int, overlayCategory: str|None):
+    def __init__(self, context: Context, vromStart: int, vromEnd: int, inFileOffset: int, vram: int, name: str, words: list[int], sectionType: FileSectionType, segmentVromStart: int, overlayCategory: str|None) -> None:
         """Constructor
 
         Args:
@@ -40,6 +40,7 @@ class ElementBase:
         self.name: str = name
         self.words: list[int] = words
         self.sectionType: FileSectionType = sectionType
+        self.customSectionName: str|None = None
 
         self.commentOffset: int = 0
         "This value is added to the first column of the disassembled line comment, allowing to change this value without messing inFileOffset"
@@ -67,15 +68,23 @@ class ElementBase:
         return self.vram + self.sizew * 4
 
 
-    def setVram(self, vram: int):
+    def setVram(self, vram: int) -> None:
         self.vram = vram
 
-    def setCommentOffset(self, commentOffset: int):
+    def setCommentOffset(self, commentOffset: int) -> None:
         self.commentOffset = commentOffset
 
+    def getName(self) -> str:
+        return self.name
+
     def getVromOffset(self, localOffset: int) -> int:
         return self.vromStart + localOffset
 
+    def getVromOffsetNone(self, localOffset: int) -> int|None:
+        if self.sectionType == FileSectionType.Bss:
+            return None
+        return self.getVromOffset(localOffset)
+
     def getVramOffset(self, localOffset: int) -> int:
         return self.vram + localOffset
 
@@ -87,12 +96,13 @@ class ElementBase:
         return True
 
 
-    def getLabelFromSymbol(self, sym: ContextSymbol|None) -> str:
+    def getLabelFromSymbol(self, sym: ContextSymbol|None, symName: str|None) -> str:
         "Generates a glabel for the passed symbol, including an optional index value if it was set and it is enabled in the GlobalConfig"
         if sym is not None:
-            label = sym.getSymbolLabel()
-            if not label:
+            label = sym.getLabelMacro(isInMiddleLabel=False)
+            if label is None:
                 return ""
+            label += f" {symName or sym.getName()}"
             if GlobalConfig.GLABEL_ASM_COUNT:
                 if self.index is not None:
                     label += f" # {self.index}"
@@ -101,7 +111,7 @@ class ElementBase:
         return ""
 
 
-    def analyze(self):
+    def analyze(self) -> None:
         """Scans the words of this element, gathering as much info as possible.
 
         This method should be called only once for each element.
@@ -202,9 +212,12 @@ class ElementBase:
         return self.context.unknownSegment
 
 
-    def getSymbol(self, vramAddress: int, tryPlusOffset: bool = True, checkUpperLimit: bool = True, checkGlobalSegment: bool = True) -> ContextSymbol|None:
+    def getSymbol(self, vramAddress: int, *, vromAddress: int|None=None, tryPlusOffset: bool=True, checkUpperLimit: bool=True, checkGlobalSegment: bool=True) -> ContextSymbol|None:
         "Searches symbol or a symbol with an addend if `tryPlusOffset` is True"
 
+        if vromAddress is not None:
+            return self.getSegmentForVrom(vromAddress).getSymbol(vramAddress, tryPlusOffset=tryPlusOffset, checkUpperLimit=checkUpperLimit)
+
         if self.overlayCategory is None or checkGlobalSegment:
             contextSym = self.context.globalSegment.getSymbol(vramAddress, tryPlusOffset=tryPlusOffset, checkUpperLimit=checkUpperLimit)
             if contextSym is not None:
@@ -245,7 +258,7 @@ class ElementBase:
             return None
         return contextSym
 
-    def getSymbolByVrom(self, vromAddress: int, tryPlusOffset: bool = True, checkUpperLimit: bool = True) -> ContextSymbol|None:
+    def getSymbolByVrom(self, vromAddress: int, *, tryPlusOffset: bool = True, checkUpperLimit: bool = True) -> ContextSymbol|None:
         segment = self.getSegmentForVrom(vromAddress)
         vram = segment.vromToVram(vromAddress)
         if vram is None:
@@ -262,35 +275,35 @@ class ElementBase:
         return segment.getConstant(constantValue)
 
 
-    def addSymbol(self, vramAddress: int, sectionType: FileSectionType=FileSectionType.Unknown, isAutogenerated: bool=False, symbolVrom: int|None=None) -> ContextSymbol:
+    def addSymbol(self, vramAddress: int, *, sectionType: FileSectionType=FileSectionType.Unknown, isAutogenerated: bool=False, symbolVrom: int|None=None, allowAddendInstead: bool=False) -> ContextSymbol:
         if symbolVrom is not None:
             segment = self.getSegmentForVrom(symbolVrom)
         else:
             segment = self.getSegmentForVram(vramAddress)
-        return segment.addSymbol(vramAddress, sectionType=sectionType, isAutogenerated=isAutogenerated, vromAddress=symbolVrom)
+        return segment.addSymbol(vramAddress, sectionType=sectionType, isAutogenerated=isAutogenerated, vromAddress=symbolVrom, allowAddendInstead=allowAddendInstead)
 
-    def addFunction(self, vramAddress: int, isAutogenerated: bool=False, symbolVrom: int|None=None) -> ContextSymbol:
+    def addFunction(self, vramAddress: int, *, isAutogenerated: bool=False, symbolVrom: int|None=None) -> ContextSymbol:
         if symbolVrom is not None:
             segment = self.getSegmentForVrom(symbolVrom)
         else:
             segment = self.getSegmentForVram(vramAddress)
         return segment.addFunction(vramAddress, isAutogenerated=isAutogenerated, vromAddress=symbolVrom)
 
-    def addBranchLabel(self, vramAddress: int, isAutogenerated: bool=False, symbolVrom: int|None=None) -> ContextSymbol:
+    def addBranchLabel(self, vramAddress: int, *, isAutogenerated: bool=False, symbolVrom: int|None=None) -> ContextSymbol:
         if symbolVrom is not None:
             segment = self.getSegmentForVrom(symbolVrom)
         else:
             segment = self.getSegmentForVram(vramAddress)
         return segment.addBranchLabel(vramAddress, isAutogenerated=isAutogenerated, vromAddress=symbolVrom)
 
-    def addJumpTable(self, vramAddress: int, isAutogenerated: bool=False, symbolVrom: int|None=None) -> ContextSymbol:
+    def addJumpTable(self, vramAddress: int, *, isAutogenerated: bool=False, symbolVrom: int|None=None) -> ContextSymbol:
         if symbolVrom is not None:
             segment = self.getSegmentForVrom(symbolVrom)
         else:
             segment = self.getSegmentForVram(vramAddress)
         return segment.addJumpTable(vramAddress, isAutogenerated=isAutogenerated, vromAddress=symbolVrom)
 
-    def addJumpTableLabel(self, vramAddress: int, isAutogenerated: bool=False, symbolVrom: int|None=None) -> ContextSymbol:
+    def addJumpTableLabel(self, vramAddress: int, *, isAutogenerated: bool=False, symbolVrom: int|None=None) -> ContextSymbol:
         if symbolVrom is not None:
             segment = self.getSegmentForVrom(symbolVrom)
         else:
@@ -298,9 +311,17 @@ class ElementBase:
         return segment.addJumpTableLabel(vramAddress, isAutogenerated=isAutogenerated, vromAddress=symbolVrom)
 
 
-    def addConstant(self, constantValue: int, name: str) -> ContextSymbol:
+    def removeSymbol(self, vramAddress: int, *, symbolVrom: int|None=None) -> None:
+        if symbolVrom is not None:
+            segment = self.getSegmentForVrom(symbolVrom)
+        else:
+            segment = self.getSegmentForVram(vramAddress)
+        return segment.removeSymbol(vramAddress)
+
+
+    def addConstant(self, constantValue: int, name: str, *, isAutogenerated: bool=False) -> ContextSymbol:
         segment = self.getSegment()
-        return segment.addConstant(constantValue, name)
+        return segment.addConstant(constantValue, name, isAutogenerated=isAutogenerated)
 
 
     def addPointerInDataReference(self, pointer: int) -> None:
@@ -316,10 +337,7 @@ class ElementBase:
         return segment.getAndPopPointerInDataReferencesRange(low, high)
 
 
-    def canUseAddendsOnData(self) -> bool:
-        segment = self.getSegmentForVram(self.vram)
-        return GlobalConfig.ALLOW_ALL_ADDENDS_ON_DATA or self.vram in segment.dataSymbolsWithReferencesWithAddends
-
-    def canUseConstantsOnData(self) -> bool:
-        segment = self.getSegmentForVram(self.vram)
-        return self.vram in segment.dataReferencingConstants
+    def getSectionName(self) -> str:
+        if self.customSectionName is not None:
+            return self.customSectionName
+        return self.sectionType.toSectionName()
