@@ -1,12 +1,13 @@
 #!/usr/bin/env python3
 
-# SPDX-FileCopyrightText: © 2022 Decompollaborate
+# SPDX-FileCopyrightText: © 2022-2024 Decompollaborate
 # SPDX-License-Identifier: MIT
 
 from __future__ import annotations
 
 import dataclasses
 import struct
+from typing import Generator
 
 from .. import common
 
@@ -29,7 +30,7 @@ class Elf32SectionHeaderEntry:
                                 # 0x28
 
     @staticmethod
-    def fromBytearray(array_of_bytes: bytearray, offset: int = 0) -> Elf32SectionHeaderEntry:
+    def fromBytearray(array_of_bytes: bytes, offset: int = 0) -> Elf32SectionHeaderEntry:
         headerFormat = common.GlobalConfig.ENDIAN.toFormatString() + "10I"
         unpacked = struct.unpack_from(headerFormat, array_of_bytes, offset)
 
@@ -37,7 +38,7 @@ class Elf32SectionHeaderEntry:
 
 
 class Elf32SectionHeaders:
-    def __init__(self, array_of_bytes: bytearray, shoff: int, shnum: int):
+    def __init__(self, array_of_bytes: bytes, shoff: int, shnum: int) -> None:
         self.sections: list[Elf32SectionHeaderEntry] = list()
         self.shoff: int = shoff
         self.shnum: int = shnum
@@ -67,6 +68,9 @@ class Elf32SectionHeaders:
             return None
         return self.sections[key]
 
-    def __iter__(self):
+    def __iter__(self) -> Generator[Elf32SectionHeaderEntry, None, None]:
         for entry in self.sections:
             yield entry
+
+    def __len__(self) -> int:
+        return len(self.sections)
