@@ -1,15 +1,17 @@
 #!/usr/bin/env python3
 
-# SPDX-FileCopyrightText: © 2022 Decompollaborate
+# SPDX-FileCopyrightText: © 2022-2024 Decompollaborate
 # SPDX-License-Identifier: MIT
 
 from __future__ import annotations
 
+from typing import Generator
+
 
 # a.k.a. strtab (string table)
 class Elf32StringTable:
-    def __init__(self, array_of_bytes: bytearray, offset: int, rawsize: int):
-        self.strings: bytearray = array_of_bytes[offset:offset+rawsize]
+    def __init__(self, array_of_bytes: bytes, offset: int, rawsize: int) -> None:
+        self.strings: bytes = array_of_bytes[offset:offset+rawsize]
         self.offset: int = offset
         self.rawsize: int = rawsize
 
@@ -23,7 +25,7 @@ class Elf32StringTable:
 
         return buffer.decode()
 
-    def __iter__(self):
+    def __iter__(self) -> Generator[str, None, None]:
         i = 0
         while i < self.rawsize:
             string = self[i]
