@@ -1,12 +1,13 @@
 #!/usr/bin/env python3
 
-# SPDX-FileCopyrightText: © 2022 Decompollaborate
+# SPDX-FileCopyrightText: © 2022-2024 Decompollaborate
 # SPDX-License-Identifier: MIT
 
 from __future__ import annotations
 
 import argparse
 from pathlib import Path
+import rabbitizer
 
 from .. import common
 from .. import elf32
@@ -18,10 +19,11 @@ from .. import __version__
 PROGNAME = "elfObjDisasm"
 
 
-def getArgsParser() -> argparse.ArgumentParser:
-    # TODO
-    description = ""
-    parser = argparse.ArgumentParser(description=description)
+def getToolDescription() -> str:
+    return "Experimental MIPS elf disassembler"
+
+def addOptionsToParser(parser: argparse.ArgumentParser) -> argparse.ArgumentParser:
+    parser.add_argument("-V", "--version", action="version", version=f"%(prog)s {__version__}")
 
     parser.add_argument("binary", help="Path to input elf binary file")
     parser.add_argument("output", help="Path to output. Use '-' to print to stdout instead")
@@ -32,12 +34,18 @@ def getArgsParser() -> argparse.ArgumentParser:
 
     parser.add_argument("--split-functions", help="Enables the function and rodata splitter. Expects a path to place the splited functions", metavar="PATH")
 
+    parser.add_argument("--instr-category", help="The instruction category to use when disassembling every passed instruction. Defaults to 'cpu'", choices=["cpu", "rsp", "r3000gte", "r4000allegrex", "r5900"])
+
+    parser.add_argument("--function-info", help="Specifies a path where to output a csvs sumary file of every analyzed function", metavar="PATH")
+
 
     readelfOptions = parser.add_argument_group("readelf-like flags")
 
-    readelfOptions.add_argument("-a", "--all", help="Equivalent to --file-header --syms --relocs --display-got", action="store_true")
+    readelfOptions.add_argument("-a", "--all", help="Equivalent to --file-header --section-headers --syms --dyn-syms --relocs --display-got", action="store_true")
     readelfOptions.add_argument("--file-header", help="Display the ELF file header", action="store_true")
+    readelfOptions.add_argument("-S", "--section-headers", help="Display the sections' header", action="store_true")
     readelfOptions.add_argument("-s", "--syms", help="Display the symbol table", action="store_true")
+    readelfOptions.add_argument("--dyn-syms", help="Display the dynamic symbol table", action="store_true")
     readelfOptions.add_argument("-r", "--relocs", help="Display the relocations (if present)", action="store_true")
     readelfOptions.add_argument("--display-got", help="Shows Global offset table information", action="store_true")
 
@@ -52,13 +60,17 @@ def getArgsParser() -> argparse.ArgumentParser:
 
     return parser
 
+def getArgsParser() -> argparse.ArgumentParser:
+    parser = argparse.ArgumentParser(description=getToolDescription(), prog=PROGNAME, formatter_class=common.Utils.PreserveWhiteSpaceWrapRawTextHelpFormatter)
+    return addOptionsToParser(parser)
+
 def applyArgs(args: argparse.Namespace) -> None:
     if args.libultra_syms is None:
         args.libultra_syms = False
     if args.hardware_regs is None:
         args.hardware_regs = False
-    mips.InstructionConfig.parseArgs(args)
     common.GlobalConfig.parseArgs(args)
+    mips.InstructionConfig.parseArgs(args)
 
 def applyGlobalConfigurations() -> None:
     common.GlobalConfig.REMOVE_POINTERS = False
@@ -67,20 +79,30 @@ def applyGlobalConfigurations() -> None:
 
     common.GlobalConfig.ALLOW_UNKSEGMENT = False
 
+    common.GlobalConfig.INPUT_FILE_TYPE = common.InputFileType.ELF
+
 
 def applyReadelfLikeFlags(elfFile: elf32.Elf32File, args: argparse.Namespace) -> None:
     if args.all:
         elfFile.readelf_fileHeader()
+        elfFile.readelf_sectionHeaders()
         elfFile.readelf_syms()
+        elfFile.readelf_dyn_syms()
         elfFile.readelf_relocs()
         elfFile.readelf_displayGot()
     else:
         if args.file_header:
             elfFile.readelf_fileHeader()
 
+        if args.section_headers:
+            elfFile.readelf_sectionHeaders()
+
         if args.syms:
             elfFile.readelf_syms()
 
+        if args.dyn_syms:
+            elfFile.readelf_dyn_syms()
+
         if args.relocs:
             elfFile.readelf_relocs()
 
@@ -90,67 +112,95 @@ def applyReadelfLikeFlags(elfFile: elf32.Elf32File, args: argparse.Namespace) ->
     if args.readelf_only:
         exit(0)
 
+SpecialSectionNames = {".text", ".data", ".rodata", ".bss"}
 
-def getOutputPath(inputPath: Path, textOutput: Path, dataOutput: Path, sectionType: common.FileSectionType) -> Path:
+def getOutputPath(inputPath: Path, textOutput: Path, dataOutput: Path, sectionType: common.FileSectionType, sectionName: str) -> Path:
     outputPath = dataOutput
     if sectionType == common.FileSectionType.Text:
         outputPath = textOutput
 
     outputFilePath = outputPath
     if str(outputPath) != "-":
-        outputFilePath = outputFilePath / inputPath.stem
+        outname = f"{inputPath.stem}"
+        if sectionName not in SpecialSectionNames:
+            outname += f"_{sectionName}"
+        outputFilePath = outputFilePath / outname
 
     return outputFilePath
 
-def getProcessedSections(context: common.Context, elfFile: elf32.Elf32File, array_of_bytes: bytearray, inputPath: Path, textOutput: Path, dataOutput: Path) -> tuple[dict[common.FileSectionType, list[mips.sections.SectionBase]], dict[common.FileSectionType, list[Path]]]:
-    processedSegments: dict[common.FileSectionType, list[mips.sections.SectionBase]] = dict()
-    segmentPaths: dict[common.FileSectionType, list[Path]] = dict()
+def processSection(
+        context: common.Context, array_of_bytes: bytes,
+        processedSections: dict[common.FileSectionType, list[mips.sections.SectionBase]],
+        segmentPaths: dict[common.FileSectionType, list[Path]],
+        sectionsPerName: dict[str, mips.sections.SectionBase],
+        inputPath: Path, textOutput: Path, dataOutput: Path,
+        sectionName: str, sectionEntry: elf32.Elf32SectionHeaderEntry,
+        sectionType: common.FileSectionType,
+        sectionClass: type[mips.sections.SectionText]|type[mips.sections.SectionData]|type[mips.sections.SectionRodata]|type[mips.sections.SectionBss]
+    ) -> None:
+    outputFilePath = getOutputPath(inputPath, textOutput, dataOutput, sectionType, sectionName)
 
-    for sectionType, sectionEntry in elfFile.progbits.items():
-        outputFilePath = getOutputPath(inputPath, textOutput, dataOutput, sectionType)
+    vromStart = sectionEntry.offset
+    vromEnd = vromStart + sectionEntry.size
+    vramStart = sectionEntry.addr
+    if sectionEntry.addr == 0:
+        vramStart = sectionEntry.offset
 
-        vromStart = sectionEntry.offset
-        vromEnd = vromStart + sectionEntry.size
-        vramStart = sectionEntry.addr
-        if sectionEntry.addr == 0:
-            vramStart = sectionEntry.offset
+    mipsSection: mips.sections.SectionBase
 
-        mipsSection: mips.sections.SectionBase
-        if sectionType == common.FileSectionType.Text:
-            mipsSection = mips.sections.SectionText(context, vromStart, vromEnd, vramStart, inputPath.stem, array_of_bytes, 0, None)
-        elif sectionType == common.FileSectionType.Data:
-            mipsSection = mips.sections.SectionData(context, vromStart, vromEnd, vramStart, inputPath.stem, array_of_bytes, 0, None)
-        elif sectionType == common.FileSectionType.Rodata:
-            mipsSection = mips.sections.SectionRodata(context, vromStart, vromEnd, vramStart, inputPath.stem, array_of_bytes, 0, None)
-        else:
-            common.Utils.eprint(f"Error! Invalid section type '{sectionType}'")
-            exit(-1)
-        mipsSection.setCommentOffset(vromStart)
-        processedSegments[sectionType] = [mipsSection]
-        segmentPaths[sectionType] = [outputFilePath]
 
-    if elfFile.nobits is not None:
-        outputFilePath = getOutputPath(inputPath, textOutput, dataOutput, common.FileSectionType.Bss)
+    inname = f"{inputPath.stem}"
+    if sectionName not in SpecialSectionNames:
+        inname += f"_{sectionName}"
 
-        vromStart = elfFile.nobits.offset
-        vromEnd = vromStart + elfFile.nobits.size
-        bssStart = elfFile.nobits.addr
-        if elfFile.nobits.addr == 0:
-            bssStart = elfFile.nobits.offset
-        bssEnd = bssStart + elfFile.nobits.size
+    if issubclass(sectionClass, mips.sections.SectionBss):
+        bssStart = vramStart
+        bssEnd = bssStart + sectionEntry.size
+        mipsSection = sectionClass(context, vromStart, vromEnd, bssStart, bssEnd, inname, 0, None)
+    else:
+        mipsSection = sectionClass(context, vromStart, vromEnd, vramStart, inname, array_of_bytes, 0, None)
 
-        mipsSection = mips.sections.SectionBss(context, vromStart, vromEnd, bssStart, bssEnd, inputPath.stem, 0, None)
-        mipsSection.setCommentOffset(vromStart)
-        processedSegments[common.FileSectionType.Bss] = [mipsSection]
-        segmentPaths[common.FileSectionType.Bss] = [outputFilePath]
+    mipsSection.setCommentOffset(vromStart)
+    mipsSection.customSectionName = sectionName
+    processedSections[sectionType].append(mipsSection)
+    segmentPaths[sectionType].append(outputFilePath)
+    sectionsPerName[sectionName] = mipsSection
 
-    return processedSegments, segmentPaths
+
+def getProcessedSections(context: common.Context, elfFile: elf32.Elf32File, array_of_bytes: bytes, inputPath: Path, textOutput: Path, dataOutput: Path) -> tuple[dict[common.FileSectionType, list[mips.sections.SectionBase]], dict[common.FileSectionType, list[Path]], dict[str, mips.sections.SectionBase]]:
+    processedSections: dict[common.FileSectionType, list[mips.sections.SectionBase]] = {
+        common.FileSectionType.Text: [],
+        common.FileSectionType.Data: [],
+        common.FileSectionType.Rodata: [],
+        common.FileSectionType.Bss: [],
+    }
+    segmentPaths: dict[common.FileSectionType, list[Path]] = {
+        common.FileSectionType.Text: [],
+        common.FileSectionType.Data: [],
+        common.FileSectionType.Rodata: [],
+        common.FileSectionType.Bss: [],
+    }
+    sectionsPerName: dict[str, mips.sections.SectionBase] = dict()
+
+    for sectionName, sectionEntry in elfFile.progbitsExecute.items():
+        processSection(context, array_of_bytes, processedSections, segmentPaths, sectionsPerName, inputPath, textOutput, dataOutput, sectionName, sectionEntry, common.FileSectionType.Text, mips.sections.SectionText)
+
+    for sectionName, sectionEntry in elfFile.progbitsWrite.items():
+        processSection(context, array_of_bytes, processedSections, segmentPaths, sectionsPerName, inputPath, textOutput, dataOutput, sectionName, sectionEntry, common.FileSectionType.Data, mips.sections.SectionData)
+
+    for sectionName, sectionEntry in elfFile.progbitsNoWrite.items():
+        processSection(context, array_of_bytes, processedSections, segmentPaths, sectionsPerName, inputPath, textOutput, dataOutput, sectionName, sectionEntry, common.FileSectionType.Rodata, mips.sections.SectionRodata)
+
+    for sectionName, sectionEntry in elfFile.nobitsPerName.items():
+        processSection(context, array_of_bytes, processedSections, segmentPaths, sectionsPerName, inputPath, textOutput, dataOutput, sectionName, sectionEntry, common.FileSectionType.Bss, mips.sections.SectionBss)
+
+    return processedSections, segmentPaths, sectionsPerName
 
 def changeGlobalSegmentRanges(context: common.Context, processedSegments: dict[common.FileSectionType, list[mips.sections.SectionBase]]) -> None:
-    lowestVromStart = None
-    highestVromEnd = None
-    lowestVramStart = None
-    highestVramEnd = None
+    lowestVromStart: int|None = None
+    highestVromEnd: int|None = None
+    lowestVramStart: int|None = None
+    highestVramEnd: int|None = None
 
     for subSegment in processedSegments.values():
         for section in subSegment:
@@ -191,10 +241,16 @@ def addContextSymFromSymEntry(context: common.Context, symEntry: elf32.Elf32SymE
     elif symEntry.stType == elf32.Elf32SymbolTableType.SECTION.value:
         # print(symEntry)
         return None
-    elif symEntry.stType == elf32.Elf32SymbolTableType.NOTYPE.value and symEntry.shndx == elf32.Elf32SectionHeaderNumber.ABS.value:
-        segment = context.globalSegment
-        contextSym = segment.addSymbol(symAddress, vromAddress=symVrom)
-        contextSym.isElfNotype = True
+    elif symEntry.stType == elf32.Elf32SymbolTableType.NOTYPE.value:
+        if symEntry.shndx == elf32.Elf32SectionHeaderNumber.ABS.value:
+            segment = context.globalSegment
+            contextSym = segment.addSymbol(symAddress, vromAddress=symVrom)
+            contextSym.isElfNotype = True
+        else:
+            bind = elf32.Elf32SymbolTableBinding.fromValue(symEntry.stBind)
+            if bind != elf32.Elf32SymbolTableBinding.LOCAL:
+                common.Utils.eprint(f"Warning: Non-LOCAL ({bind}) NOTYPE symbol '{symName}' has an unhandled shndx value: '0x{symEntry.shndx:X}'")
+            contextSym = segment.addSymbol(symAddress, vromAddress=symVrom)
     else:
         common.Utils.eprint(f"Warning: symbol '{symName}' has an unhandled stType: '{symEntry.stType}'")
         contextSym = segment.addSymbol(symAddress, vromAddress=symVrom)
@@ -210,7 +266,7 @@ def addContextSymFromSymEntry(context: common.Context, symEntry: elf32.Elf32SymE
     return contextSym
 
 
-def insertSymtabIntoContext(context: common.Context, symbolTable: elf32.Elf32Syms, stringTable: elf32.Elf32StringTable, elfFile: elf32.Elf32File):
+def insertSymtabIntoContext(context: common.Context, symbolTable: elf32.Elf32Syms, stringTable: elf32.Elf32StringTable, elfFile: elf32.Elf32File) -> None:
     # Use the symbol table to replace symbol names present in disassembled sections
     for i, symEntry in enumerate(symbolTable):
         symName = stringTable[symEntry.name]
@@ -240,7 +296,7 @@ def insertSymtabIntoContext(context: common.Context, symbolTable: elf32.Elf32Sym
         if contextSym is not None:
             contextSym.sectionType = sectType
 
-def insertDynsymIntoContext(context: common.Context, symbolTable: elf32.Elf32Syms, stringTable: elf32.Elf32StringTable):
+def insertDynsymIntoContext(context: common.Context, symbolTable: elf32.Elf32Syms, stringTable: elf32.Elf32StringTable) -> None:
     for symEntry in symbolTable:
         if symEntry.value == 0 or symEntry.shndx == 0:
             continue
@@ -248,13 +304,12 @@ def insertDynsymIntoContext(context: common.Context, symbolTable: elf32.Elf32Sym
         symName = stringTable[symEntry.name]
         addContextSymFromSymEntry(context, symEntry, symEntry.value, symName)
 
-def insertGotIntoContext(context: common.Context, got: elf32.Elf32GlobalOffsetTable, stringTable: elf32.Elf32StringTable):
+def insertGotIntoContext(context: common.Context, got: elf32.Elf32GlobalOffsetTable, stringTable: elf32.Elf32StringTable) -> None:
     lazyResolver = got.localsTable[0]
-    contextSym: common.ContextSymbol|None
-    contextSym = context.globalSegment.addSymbol(lazyResolver)
-    contextSym.name = f"$$.LazyResolver"
-    contextSym.isUserDeclared = True
-    contextSym.isGotLocal = True
+    lazyResolverSym = context.globalSegment.addSymbol(lazyResolver)
+    lazyResolverSym.name = f"$$.LazyResolver"
+    lazyResolverSym.isUserDeclared = True
+    lazyResolverSym.isGotLocal = True
 
     gotIndex = len(got.localsTable)
 
@@ -267,27 +322,34 @@ def insertGotIntoContext(context: common.Context, got: elf32.Elf32GlobalOffsetTa
         gotIndex += 1
 
 
-def injectAllElfSymbols(context: common.Context, elfFile: elf32.Elf32File, processedSegments: dict[common.FileSectionType, list[mips.sections.SectionBase]]) -> None:
+def injectAllElfSymbols(context: common.Context, elfFile: elf32.Elf32File, processedSegments: dict[common.FileSectionType, list[mips.sections.SectionBase]], sectionsPerName: dict[str, mips.sections.SectionBase]) -> None:
     if elfFile.symtab is not None and elfFile.strtab is not None:
         # Inject symbols from the reloc table referenced in each section
         if elfFile.header.type == elf32.Elf32ObjectFileType.REL.value:
-            for sectType, relocs in elfFile.rel.items():
+            for sectionName, relocs in elfFile.relPerName.items():
+                subSegment = sectionsPerName.get(sectionName, None)
                 for rel in relocs:
                     symbolEntry = elfFile.symtab[rel.rSym]
                     symbolName = elfFile.strtab[symbolEntry.name]
-                    if symbolName == "":
+
+                    if symbolEntry.stType != elf32.Elf32SymbolTableType.SECTION.value:
+                        if symbolName == "":
+                            continue
+
+                    if subSegment is None:
+                        common.Utils.eprint(f"Warning: reloc '{rel}' references unhandled section '{sectionName}'")
                         continue
 
-                    subSegment = processedSegments[sectType][0]
-
                     relocVrom = subSegment.vromStart + rel.offset
-                    relocInfo = common.RelocationInfo(common.RelocType(rel.rType), symbolName)
+                    relocInfo = context.addGlobalReloc(relocVrom, common.RelocType(rel.rType), symbolName)
                     if symbolEntry.stType == elf32.Elf32SymbolTableType.SECTION.value:
-                        sectionType = common.FileSectionType.fromStr(symbolName)
-                        sectionVram = processedSegments[sectionType][0].vram
-                        relocInfo.staticReference = common.RelocationStaticReference(sectionType, sectionVram)
+                        sectionEntry = elfFile.sectionHeaders[symbolEntry.shndx]
+                        assert sectionEntry is not None, rel
 
-                    context.globalRelocationOverrides[relocVrom] = relocInfo
+                        sectionType = common.FileSectionType.fromStr(elfFile.shstrtab[sectionEntry.name])
+                        if sectionType != common.FileSectionType.Invalid:
+                            sectionVram = processedSegments[sectionType][0].vram
+                            relocInfo.staticReference = common.RelocationStaticReference(sectionType, sectionVram)
 
         # Use the symtab to replace symbol names present in disassembled sections
         insertSymtabIntoContext(context, elfFile.symtab, elfFile.strtab, elfFile)
@@ -301,18 +363,24 @@ def injectAllElfSymbols(context: common.Context, elfFile: elf32.Elf32File, proce
     return
 
 def processGlobalOffsetTable(context: common.Context, elfFile: elf32.Elf32File) -> None:
-    if elfFile.dynamic is not None and elfFile.got is not None:
+    if elfFile.dynamic is not None:
         common.GlobalConfig.GP_VALUE = elfFile.dynamic.getGpValue()
 
+    if elfFile.reginfo is not None:
+        common.GlobalConfig.GP_VALUE = elfFile.reginfo.gpValue
+
+    if elfFile.dynamic is not None and elfFile.got is not None:
         globalsTable = [gotEntry.getAddress() for gotEntry in elfFile.got.globalsTable]
 
         assert elfFile.dynamic.pltGot is not None
         context.initGotTable(elfFile.dynamic.pltGot, elfFile.got.localsTable, globalsTable)
+
+        for small in elfFile.smallSections.values():
+            context.addSmallSection(small.addr, small.size)
     return
 
 
-def elfObjDisasmMain():
-    args = getArgsParser().parse_args()
+def processArguments(args: argparse.Namespace) -> int:
     applyArgs(args)
 
     applyGlobalConfigurations()
@@ -329,6 +397,14 @@ def elfObjDisasmMain():
     elfFile.handleHeaderIdent()
     elfFile.handleFlags()
 
+    instrCategory = args.instr_category
+    if instrCategory is None:
+        if elf32.Elf32Constants.Elf32HeaderFlag._5900 in elfFile.elfFlags:
+            instrCategory = "r5900"
+
+    common.Utils.printQuietless(f"{PROGNAME} {inputPath}: Processing global offset table...")
+    processGlobalOffsetTable(context, elfFile)
+
     applyReadelfLikeFlags(elfFile, args)
 
     textOutput = Path(args.output)
@@ -342,16 +418,27 @@ def elfObjDisasmMain():
         splits = common.FileSplitFormat()
         splits.readCsvFile(Path(args.file_splits))
         processedSegments, segmentPaths = fec.FrontendUtilities.getSplittedSections(context, splits, array_of_bytes, inputPath, textOutput, dataOutput)
+        sectionsPerName: dict[str, mips.sections.SectionBase] = dict()
+        for sectType, subSection in processedSegments.items():
+            if len(subSection) < 1:
+                continue
+            if sectType == common.FileSectionType.Text:
+                sectionsPerName[".text"] = subSection[0]
+            elif sectType == common.FileSectionType.Data:
+                sectionsPerName[".data"] = subSection[0]
+            elif sectType == common.FileSectionType.Rodata:
+                sectionsPerName[".rodata"] = subSection[0]
+            elif sectType == common.FileSectionType.Bss:
+                sectionsPerName[".bss"] = subSection[0]
     else:
-        processedSegments, segmentPaths = getProcessedSections(context, elfFile, array_of_bytes, inputPath, textOutput, dataOutput)
+        processedSegments, segmentPaths, sectionsPerName = getProcessedSections(context, elfFile, array_of_bytes, inputPath, textOutput, dataOutput)
 
     changeGlobalSegmentRanges(context, processedSegments)
 
-    common.Utils.printQuietless(f"{PROGNAME} {inputPath}: Processing global offset table...")
-    processGlobalOffsetTable(context, elfFile)
+    fec.FrontendUtilities.configureProcessedFiles(processedSegments, instrCategory)
 
     common.Utils.printQuietless(f"{PROGNAME} {inputPath}: Injecting elf symbols...")
-    injectAllElfSymbols(context, elfFile, processedSegments)
+    injectAllElfSymbols(context, elfFile, processedSegments, sectionsPerName)
 
     processedFilesCount = 0
     for sect in processedSegments.values():
@@ -377,4 +464,22 @@ def elfObjDisasmMain():
         contextPath.parent.mkdir(parents=True, exist_ok=True)
         context.saveContextToFile(contextPath)
 
+    if args.function_info is not None:
+        fec.FrontendUtilities.writeFunctionInfoCsv(processedSegments, Path(args.function_info))
+
     common.Utils.printQuietless(f"{PROGNAME} {inputPath}: Done!")
+
+    return 0
+
+def addSubparser(subparser: argparse._SubParsersAction[argparse.ArgumentParser]) -> None:
+    parser = subparser.add_parser(PROGNAME, help=getToolDescription(), formatter_class=common.Utils.PreserveWhiteSpaceWrapRawTextHelpFormatter)
+
+    addOptionsToParser(parser)
+
+    parser.set_defaults(func=processArguments)
+
+
+def elfObjDisasmMain() -> int:
+    args = getArgsParser().parse_args()
+
+    return processArguments(args)
