@@ -1,6 +1,6 @@
 #!/usr/bin/env python3
 
-# SPDX-FileCopyrightText: © 2022 Decompollaborate
+# SPDX-FileCopyrightText: © 2022-2024 Decompollaborate
 # SPDX-License-Identifier: MIT
 
 from __future__ import annotations
@@ -14,9 +14,10 @@ from .. import common
 
 from . import sections
 from . import symbols
+from . import FunctionRodataEntry
 
 
-def createSectionFromSplitEntry(splitEntry: common.FileSplitEntry, array_of_bytes: bytearray, outputPath: Path, context: common.Context) -> sections.SectionBase:
+def createSectionFromSplitEntry(splitEntry: common.FileSplitEntry, array_of_bytes: bytes, context: common.Context, vromStart: int = 0) -> sections.SectionBase:
     offsetStart = splitEntry.offset
     offsetEnd = splitEntry.nextOffset
 
@@ -30,20 +31,26 @@ def createSectionFromSplitEntry(splitEntry: common.FileSplitEntry, array_of_byte
     elif offsetStart >= 0:
         common.Utils.printVerbose(f"Parsing since offset 0x{offsetStart:02X}")
 
+    sectionStart = vromStart + offsetStart
+    sectionEnd = vromStart + offsetEnd
+
     common.Utils.printVerbose(f"Using VRAM {splitEntry.vram:08X}")
     vram = splitEntry.vram
 
     f: sections.SectionBase
     if splitEntry.section == common.FileSectionType.Text:
-        f = sections.SectionText(context, offsetStart, offsetEnd, vram, outputPath.stem, array_of_bytes, 0, None)
+        f = sections.SectionText(context, sectionStart, sectionEnd, vram, splitEntry.fileName, array_of_bytes, 0, None)
         if splitEntry.isRsp:
             f.instrCat = rabbitizer.InstrCategory.RSP
     elif splitEntry.section == common.FileSectionType.Data:
-        f = sections.SectionData(context, offsetStart, offsetEnd, vram, outputPath.stem, array_of_bytes, 0, None)
+        f = sections.SectionData(context, sectionStart, sectionEnd, vram, splitEntry.fileName, array_of_bytes, 0, None)
     elif splitEntry.section == common.FileSectionType.Rodata:
-        f = sections.SectionRodata(context, offsetStart, offsetEnd, vram, outputPath.stem, array_of_bytes, 0, None)
+        f = sections.SectionRodata(context, sectionStart, sectionEnd, vram, splitEntry.fileName, array_of_bytes, 0, None)
     elif splitEntry.section == common.FileSectionType.Bss:
-        f = sections.SectionBss(context, offsetStart, offsetEnd, splitEntry.vram, splitEntry.vram + offsetEnd - offsetStart, outputPath.stem, 0, None)
+        bssVramEnd = splitEntry.vram + offsetEnd - offsetStart
+        f = sections.SectionBss(context, sectionStart, sectionEnd, splitEntry.vram, bssVramEnd, splitEntry.fileName, 0, None)
+    elif splitEntry.section == common.FileSectionType.Reloc:
+        f = sections.SectionRelocZ64(context, sectionStart, sectionEnd, vram, splitEntry.fileName, array_of_bytes, 0, None)
     else:
         common.Utils.eprint("Error! Section not set!")
         exit(-1)
@@ -52,91 +59,47 @@ def createSectionFromSplitEntry(splitEntry: common.FileSplitEntry, array_of_byte
 
     return f
 
-def writeSection(path: Path, fileSection: sections.SectionBase):
+def writeSection(path: Path, fileSection: sections.SectionBase) -> Path:
     path.parent.mkdir(parents=True, exist_ok=True)
     fileSection.saveToFile(str(path))
     return path
 
 
+#! @deprecated: Use the static methods from FunctionRodataEntry instead
 def getRdataAndLateRodataForFunctionFromSection(func: symbols.SymbolFunction, rodataSection: sections.SectionRodata) -> tuple[list[symbols.SymbolBase], list[symbols.SymbolBase], int]:
-    rdataList: list[symbols.SymbolBase] = []
-    lateRodataList: list[symbols.SymbolBase] = []
+    entry = FunctionRodataEntry.getEntryForFuncFromSection(func, rodataSection)
+
     lateRodataSize = 0
+    for rodataSym in entry.lateRodataSyms:
+        lateRodataSize += rodataSym.sizew
 
-    intersection = func.instrAnalyzer.referencedVrams & rodataSection.symbolsVRams
-    for rodataSym in rodataSection.symbolList:
-        if rodataSym.vram not in intersection:
-            continue
-
-        if not rodataSym.shouldMigrate():
-            continue
-
-        if rodataSym.contextSym.isLateRodata() and common.GlobalConfig.COMPILER == common.Compiler.IDO:
-            lateRodataList.append(rodataSym)
-            lateRodataSize += rodataSym.sizew
-        else:
-            rdataList.append(rodataSym)
-
-    return rdataList, lateRodataList, lateRodataSize
+    return entry.rodataSyms, entry.lateRodataSyms, lateRodataSize
 
+#! @deprecated: Use the static methods from FunctionRodataEntry instead
 def getRdataAndLateRodataForFunction(func: symbols.SymbolFunction, rodataFileList: list[sections.SectionBase]) -> tuple[list[symbols.SymbolBase], list[symbols.SymbolBase], int]:
-    rdataList: list[symbols.SymbolBase] = []
-    lateRodataList: list[symbols.SymbolBase] = []
+    entry = FunctionRodataEntry.getEntryForFuncFromPossibleRodataSections(func, rodataFileList)
+
     lateRodataSize = 0
+    for rodataSym in entry.lateRodataSyms:
+        lateRodataSize += rodataSym.sizew
 
-    for rodataSection in rodataFileList:
-        assert isinstance(rodataSection, sections.SectionRodata)
+    return entry.rodataSyms, entry.lateRodataSyms, lateRodataSize
 
-        if len(rdataList) > 0 or len(lateRodataList) > 0:
-            # We already have the rodata for this function. Stop searching
-            break
+#! @deprecated: Use the static methods from FunctionRodataEntry instead
+def writeFunctionRodataToFile(f: TextIO, func: symbols.SymbolFunction, rdataList: list[symbols.SymbolBase], lateRodataList: list[symbols.SymbolBase], lateRodataSize: int=0) -> None:
+    entry = FunctionRodataEntry(func, rdataList, lateRodataList)
+    entry.writeToFile(f, writeFunction=False)
 
-        # Skip the file if there's nothing in this file refenced by the current function
-        intersection = func.instrAnalyzer.referencedVrams & rodataSection.symbolsVRams
-        if len(intersection) == 0:
-            continue
-
-        rdataList, lateRodataList, lateRodataSize = getRdataAndLateRodataForFunctionFromSection(func, rodataSection)
-
-    return rdataList, lateRodataList, lateRodataSize
-
-def writeFunctionRodataToFile(f: TextIO, func: symbols.SymbolFunction, rdataList: list[symbols.SymbolBase], lateRodataList: list[symbols.SymbolBase], lateRodataSize: int):
-    if len(rdataList) > 0:
-        # Write the rdata
-        sectionName = ".rodata"
-        f.write(f".section {sectionName}" + common.GlobalConfig.LINE_ENDS)
-        for sym in rdataList:
-            f.write(sym.disassemble(migrate=True, useGlobalLabel=True))
-            f.write(common.GlobalConfig.LINE_ENDS)
-
-    if len(lateRodataList) > 0:
-        # Write the late_rodata
-        f.write(".section .late_rodata" + common.GlobalConfig.LINE_ENDS)
-        if lateRodataSize / len(func.instructions) > 1/3:
-            align = 4
-            firstLateRodataVram = lateRodataList[0].vram
-            if firstLateRodataVram is not None and firstLateRodataVram % 8 == 0:
-                align = 8
-            f.write(f".late_rodata_alignment {align}" + common.GlobalConfig.LINE_ENDS)
-        for sym in lateRodataList:
-            f.write(sym.disassemble(migrate=True, useGlobalLabel=True))
-            f.write(common.GlobalConfig.LINE_ENDS)
-
-    if len(rdataList) > 0 or len(lateRodataList) > 0:
-        f.write(common.GlobalConfig.LINE_ENDS + ".section .text" + common.GlobalConfig.LINE_ENDS)
-
-def writeSplitedFunction(path: Path, func: symbols.SymbolFunction, rodataFileList: list[sections.SectionBase]):
+#! @deprecated: Use the static methods from FunctionRodataEntry instead
+def writeSplitedFunction(path: Path, func: symbols.SymbolFunction, rodataFileList: list[sections.SectionBase]) -> None:
     path.mkdir(parents=True, exist_ok=True)
+    entry = FunctionRodataEntry.getEntryForFuncFromPossibleRodataSections(func, rodataFileList)
 
     funcPath = path / (func.getName()+ ".s")
     with funcPath.open("w") as f:
-        rdataList, lateRodataList, lateRodataSize = getRdataAndLateRodataForFunction(func, rodataFileList)
-        writeFunctionRodataToFile(f, func, rdataList, lateRodataList, lateRodataSize)
+        entry.writeToFile(f, writeFunction=True)
 
-        # Write the function itself
-        f.write(func.disassemble(migrate=True))
-
-def writeOtherRodata(path: Path, rodataFileList: list[sections.SectionBase]):
+def writeOtherRodata(path: Path, rodataFileList: list[sections.SectionBase]) -> None:
     for rodataSection in rodataFileList:
         assert isinstance(rodataSection, sections.SectionRodata)
 
@@ -148,6 +111,8 @@ def writeOtherRodata(path: Path, rodataFileList: list[sections.SectionBase]):
                 continue
 
             rodataSymbolPath = rodataPath / (rodataSym.getName() + ".s")
+            common.Utils.printVerbose(f"Writing unmigrated rodata {rodataSymbolPath}")
+
             with rodataSymbolPath.open("w") as f:
                 f.write(".section .rodata" + common.GlobalConfig.LINE_ENDS)
                 f.write(rodataSym.disassemble(migrate=True))
