@@ -1,6 +1,6 @@
 #!/usr/bin/env python3
 
-# SPDX-FileCopyrightText: © 2022 Decompollaborate
+# SPDX-FileCopyrightText: © 2022-2024 Decompollaborate
 # SPDX-License-Identifier: MIT
 
 from __future__ import annotations
@@ -23,7 +23,7 @@ class SymbolTypeInfo:
         return self.accessType == other.accessType and self.unsignedMemoryAccess == other.unsignedMemoryAccess
 
     # https://stackoverflow.com/a/56915493/6292472
-    def __hash__(self):
+    def __hash__(self) -> int:
         return hash((self.accessType, self.unsignedMemoryAccess))
 
 
@@ -34,10 +34,19 @@ class CploadInfo:
     adduOffset: int|None = None
     reg: rabbitizer.Enum|None = None
 
+@dataclasses.dataclass
+class GpSetInfo:
+    """Info for tracking when a $gp register is set"""
+    hiOffset: int
+    loOffset: int
+    value: int
+
 
 class InstrAnalyzer:
-    def __init__(self, funcVram: int) -> None:
+    def __init__(self, funcVram: int, context: common.Context) -> None:
         self.funcVram = funcVram
+        self.context = context
+        "read-only"
 
         self.referencedVrams: set[int] = set()
         "Every referenced vram found"
@@ -59,13 +68,13 @@ class InstrAnalyzer:
         self.funcCallInstrOffsets: dict[int, int] = dict()
         "key: func call instruction offset, value: target vram"
         self.funcCallOutsideRangesOffsets: dict[int, int] = dict()
-        "key: func call instruction offset, value: target vram which is outside the [0x80000000, 0x84000000] range"
+        "key: func call instruction offset, value: target vram which is outside the known vram address range"
 
         # Jump register (jumptables)
         self.jumpRegisterIntrOffset: dict[int, int] = dict()
         self.referencedJumpTableOffsets: dict[int, int] = dict()
 
-        # Jump register (jumptables)
+        # Jump and link (functions)
         self.indirectFunctionCallIntrOffset: dict[int, int] = dict()
         self.indirectFunctionCallOffsets: dict[int, int] = dict()
 
@@ -110,6 +119,11 @@ class InstrAnalyzer:
         self.cploads: dict[int, CploadInfo] = dict()
         "Completed cpload, key: offset of last instruction of the cpload"
 
+        self.gpSetsOffsets: set[int] = set()
+        "Offsets of every instruction that set the $gp register"
+        self.gpSets: dict[int, GpSetInfo] = dict()
+        "Instructions setting the $gp register, key: offset of the low instruction"
+
 
     def processBranch(self, instr: rabbitizer.Instruction, instrOffset: int, currentVram: int) -> None:
         if instrOffset in self.branchInstrOffsets:
@@ -133,7 +147,7 @@ class InstrAnalyzer:
             return
 
         target = instr.getInstrIndexAsVram()
-        if target >= 0x84000000 or target < 0x80000000:
+        if not self.context.isAddressInGlobalRange(target):
             self.funcCallOutsideRangesOffsets[instrOffset] = target
 
         if not common.GlobalConfig.PIC:
@@ -181,11 +195,11 @@ class InstrAnalyzer:
                         if hiValue != otherLuiInstr.getProcessedImmediate() << 16:
                             return None
 
-            if common.GlobalConfig.COMPILER == common.Compiler.IDO:
+            if not common.GlobalConfig.COMPILER.value.pairMultipleHiToSameLow:
                 # IDO does not pair multiples %hi to the same %lo
                 return self.symbolLoInstrOffset[lowerOffset]
 
-            elif common.GlobalConfig.COMPILER in {common.Compiler.GCC, common.Compiler.SN64, common.Compiler.PSYQ}:
+            else:
                 if luiOffset is None or hiValue is None:
                     return None
 
@@ -234,9 +248,22 @@ class InstrAnalyzer:
                 return None
 
         # filter out stuff that may not be a real symbol
-        filterOut = common.GlobalConfig.SYMBOL_FINDER_FILTER_LOW_ADDRESSES and address < 0x80000000
-        filterOut |= common.GlobalConfig.SYMBOL_FINDER_FILTER_HIGH_ADDRESSES and address >= 0xC0000000
-        if filterOut and lowerInstr.uniqueId != rabbitizer.InstrId.cpu_addiu:
+        filterOut = False
+        if not self.context.totalVramRange.isInRange(address):
+            if common.GlobalConfig.SYMBOL_FINDER_FILTER_LOW_ADDRESSES or common.GlobalConfig.SYMBOL_FINDER_FILTER_HIGH_ADDRESSES:
+                filterOut |= common.GlobalConfig.SYMBOL_FINDER_FILTER_LOW_ADDRESSES and address < common.GlobalConfig.SYMBOL_FINDER_FILTER_ADDRESSES_ADDR_LOW
+                filterOut |= common.GlobalConfig.SYMBOL_FINDER_FILTER_HIGH_ADDRESSES and address >= common.GlobalConfig.SYMBOL_FINDER_FILTER_ADDRESSES_ADDR_HIGH
+            else:
+                filterOut |= True
+
+        if filterOut:
+            contextSym = self.context.globalSegment.getSymbol(address)
+            if contextSym is not None:
+                if contextSym.isUserDeclared:
+                    # If the user declared a symbol outside the total vram range then use it anyways
+                    filterOut = False
+
+        if address > 0 and filterOut and lowerInstr.uniqueId != rabbitizer.InstrId.cpu_addiu:
             if common.GlobalConfig.SYMBOL_FINDER_FILTERED_ADDRESSES_AS_CONSTANTS:
                 # Let's pretend this value is a constant
                 constant = address
@@ -352,13 +379,19 @@ class InstrAnalyzer:
 
         if luiOffset is not None:
             luiInstr = self.luiInstrs.get(luiOffset)
-            if luiInstr is not None:
-                if luiInstr.rt in {rabbitizer.RegGprO32.gp, rabbitizer.RegGprN32.gp}:
-                    if instr.rs in {rabbitizer.RegGprO32.gp, rabbitizer.RegGprN32.gp} and instr.rt in {rabbitizer.RegGprO32.gp, rabbitizer.RegGprN32.gp}:
+            if luiInstr is not None and luiInstr.rt in {rabbitizer.RegGprO32.gp, rabbitizer.RegGprN32.gp}:
+                if instr.readsRs() and instr.rs in {rabbitizer.RegGprO32.gp, rabbitizer.RegGprN32.gp} and instr.modifiesRt() and instr.rt in {rabbitizer.RegGprO32.gp, rabbitizer.RegGprN32.gp}:
+                    if common.GlobalConfig.PIC:
                         # cpload
                         self.unpairedCploads.append(CploadInfo(luiOffset, instrOffset))
-                        # early return to avoid counting this pairing as a symbol
-                        return
+                    else:
+                        hiGpValue = luiInstr.getProcessedImmediate() << 16
+                        loGpValue = instr.getProcessedImmediate()
+                        self.gpSets[instrOffset] = GpSetInfo(luiOffset, instrOffset, hiGpValue+loGpValue)
+                        self.gpSetsOffsets.add(luiOffset)
+                        self.gpSetsOffsets.add(instrOffset)
+                    # early return to avoid counting this pairing as a normal symbol
+                    return
 
         address = self.pairHiLo(upperHalf, luiOffset, instr, instrOffset)
         if address is None:
@@ -434,7 +467,7 @@ class InstrAnalyzer:
 
 
 
-    def printAnalisisDebugInfo_IterInfo(self, regsTracker: rabbitizer.RegistersTracker, instr: rabbitizer.Instruction, currentVram: int):
+    def printAnalisisDebugInfo_IterInfo(self, regsTracker: rabbitizer.RegistersTracker, instr: rabbitizer.Instruction, currentVram: int) -> None:
         if not common.GlobalConfig.PRINT_FUNCTION_ANALYSIS_DEBUG_INFO:
             return
 
@@ -451,7 +484,7 @@ class InstrAnalyzer:
         # print({instr.getRegisterName(register_t): f"{state_t.value:X},{state_t.loOffset:X},{state_t.dereferenced}" for register_t, state_t in regsTracker.registers.items() if state_t.hasLoValue})
         print()
 
-    def printSymbolFinderDebugInfo_UnpairedLuis(self):
+    def printSymbolFinderDebugInfo_UnpairedLuis(self) -> None:
         if not common.GlobalConfig.PRINT_UNPAIRED_LUIS_DEBUG_INFO:
             return
 
